#include "math.h"
#include "tcp-dtn.h"
#include "tcp-dtn-sink.h"
#include "dtn.h"

int hdr_dtn::offset_;

static class DTNHeaderClass : public PacketHeaderClass {
public:
        DTNHeaderClass() : PacketHeaderClass("PacketHeader/DTN",
					     sizeof(hdr_dtn)) {
		bind_offset(&hdr_dtn::offset_);
	}
} class_dtnhdr;

/*---------------------------------------------- Bundle ---------------------------------------------------*/

Bundle::Bundle()
{
	InitBundleVariables();
}

Bundle::~Bundle()
{

}

void Bundle::InitLocalVariables()
{
	nOriginalOffset_	= HeaderDTN().FragmentOffset();
	nOriginalSize_		= HeaderDTN().BundleSize();
	nReceivedBytes_		= HeaderDTN().Payload();
	pNextBundle_		= NULL;
}

void Bundle::InitBundleVariables()
{
	HeaderDTN().nSequenceNumber_		= 0;
	HeaderDTN().nSourceAddress_		= 0;
	HeaderDTN().fTimestamp_			= 0;
	HeaderDTN().nFragmentOffset_		= 0;
	HeaderDTN().nBundleSize_		= 0;
	HeaderDTN().nADUSize_			= 0;
	HeaderDTN().nAvailableBufferSpace_	= 0;
	HeaderDTN().nPayload_			= 0;
	HeaderDTN().nNextRVms_			= 0;
	HeaderDTN().bIsStatusReport_		= 0;
	HeaderDTN().bCustodyAccepted_		= 0;
	
	nOriginalOffset_ 	= 0;
	nOriginalSize_ 		= 0;
	nTCPSequenceNo_ 	= 0;
	nReceivedBytes_ 	= 0;
	fRetransmitTime_	= 0;
	bForwarded_ 		= 0;
	bAcked_			= 0;
	bSuspended_		= 0;
	pDownstreamCon_		= NULL;
	pUpstreamCon_		= NULL;
	
	pNextBundle_		= NULL;
}

bool Bundle::IsSameAs(Bundle otherBundle)
{
	if (HeaderDTN().SourceAddress() == otherBundle.HeaderDTN().SourceAddress() &&
		HeaderDTN().Timestamp() == otherBundle.HeaderDTN().Timestamp() &&
		HeaderDTN().FragmentOffset() == otherBundle.HeaderDTN().FragmentOffset() &&
		HeaderDTN().BundleSize() == otherBundle.HeaderDTN().BundleSize())
		return true;
	else
		return false;
}

bool Bundle::IsInSameADUAs(Bundle otherBundle)
{
	return (HeaderDTN().SourceAddress() == otherBundle.HeaderDTN().SourceAddress() &&
			HeaderDTN().Timestamp() == otherBundle.HeaderDTN().Timestamp());
}

/* Checks if the first bundle should precede the second bundle in the bundle list. */
bool Bundle::Precedes(Bundle otherBundle)
{
	if (HeaderDTN().SourceAddress() < otherBundle.HeaderDTN().SourceAddress())
		return true;
	
	if (HeaderDTN().SourceAddress() == otherBundle.HeaderDTN().SourceAddress() &&
		HeaderDTN().Timestamp() < otherBundle.HeaderDTN().Timestamp())
		return true;
	
	if (HeaderDTN().SourceAddress() == otherBundle.HeaderDTN().SourceAddress() &&
		HeaderDTN().Timestamp() == otherBundle.HeaderDTN().Timestamp() &&
		HeaderDTN().FragmentOffset() < otherBundle.HeaderDTN().FragmentOffset())
		return true;

	return false;
}

bool Bundle::IsCompleteADU()
{
	// printf("IsComplete: Offset = %d, Received = %d, BundleSize = %d, ADUSize = %d\n", headerDTN_.FragmentOffset(), nReceivedBytes_, headerDTN_.BundleSize(), headerDTN_.ADUSize());
	
	bool bBundleComplete = (HeaderDTN().FragmentOffset() == 0 && 
				HeaderDTN().BundleSize() == HeaderDTN().ADUSize() && 
				nReceivedBytes_ == HeaderDTN().BundleSize());
							
	return bBundleComplete;
}

int Bundle::LastBundleByte()
{
	return HeaderDTN().FragmentOffset() + HeaderDTN().BundleSize() - 1;
}

int Bundle::LastReceivedByte()
{
	return HeaderDTN().FragmentOffset() + nReceivedBytes_ - 1;
}

bool Bundle::CanBeMergedWith(Bundle otherBundle)
{
	// printf("Bundle can be merged with nReceivedBytes_1 = %d, BundleSize1 = %d, nReceivedBytes_2 = %d, BundleSize2 = %d, \n", nReceivedBytes_, 
	//	HeaderDTN().BundleSize(), otherBundle.ReceivedBytes(),  otherBundle.HeaderDTN().BundleSize());
		
	/* Fragments cannot be merged unless both are fully received. */
	if (!IsInSameADUAs(otherBundle) ||
		nReceivedBytes_ != HeaderDTN().BundleSize() ||
		otherBundle.ReceivedBytes() != otherBundle.HeaderDTN().BundleSize())
		return false;
	
	if ( (otherBundle.HeaderDTN().FragmentOffset() > LastBundleByte() + 1) || 
		(otherBundle.LastBundleByte() < HeaderDTN().FragmentOffset() - 1) )
		return false;
	else
		return true;

}

bool Bundle::MergeWith(Bundle otherBundle)
{
	/* Merge only if feasible. */
	if (!CanBeMergedWith(otherBundle))
		return false;

	// printf("Merging fragments with offsets %d and %d\n", pFirstBundleData->m_nFragmentOffset, pSecondBundleData->m_nFragmentOffset);

	if (otherBundle.LastBundleByte() > LastBundleByte())
		HeaderDTN().BundleSize() += otherBundle.LastBundleByte() - LastBundleByte();
	
	if (otherBundle.headerDTN_.FragmentOffset() < otherBundle.headerDTN_.FragmentOffset())
	{
		HeaderDTN().BundleSize() += HeaderDTN().FragmentOffset() - otherBundle.HeaderDTN().FragmentOffset();
		HeaderDTN().FragmentOffset() = otherBundle.HeaderDTN().FragmentOffset();
	}
	
	nReceivedBytes_ = headerDTN_.BundleSize();
	return true;
}

bool Bundle::FragmentReceivedBytes()
{
	if (nReceivedBytes_ > 0 && nReceivedBytes_ < HeaderDTN().BundleSize())
	{
		HeaderDTN().BundleSize() = nReceivedBytes_;
		return true;
	}
	else
		return false;
}

bool Bundle::FragmentAt(int nOffset)
{
	if (nOffset > HeaderDTN().FragmentOffset() && nOffset <= LastBundleByte())
	{
		HeaderDTN().BundleSize() = LastBundleByte() - nOffset + 1;
		HeaderDTN().FragmentOffset() = nOffset;
		return true;
	}
	else
		return false;
}

bool Bundle::FullyContains(Bundle otherBundle)
{
	bool bSameADU = IsInSameADUAs(otherBundle);
	bool bContains = (HeaderDTN().FragmentOffset() <= otherBundle.HeaderDTN().FragmentOffset() && 
			LastBundleByte() >= otherBundle.LastBundleByte());
			
	return bSameADU && bContains;
}

bool Bundle::OriginalFullyContains(Bundle otherBundle)
{
	bool bSameADU = IsInSameADUAs(otherBundle);
	bool bContains = OriginalOffset() <= otherBundle.HeaderDTN().FragmentOffset() &&
			(OriginalOffset() + OriginalSize() >= otherBundle.HeaderDTN().FragmentOffset() + otherBundle.HeaderDTN().BundleSize());
	return bSameADU && bContains;
}

bool Bundle::IsFullyReceived()
{
	return Acked() || (HeaderDTN().BundleSize() == nReceivedBytes_);
}

void Bundle::RejectCustody()
{
	HeaderDTN().IsStatusReport()	= 1;
	HeaderDTN().CustodyRejected()	= 1;
}

void Bundle::AcceptCustody()
{
	HeaderDTN().IsStatusReport()	= 1;
	HeaderDTN().CustodyAccepted()	= 1;
}

void Bundle::Dump()
{
	printf("Seq = %d, TCPSeq = %d, Offset = %d, Size = %d, RecBytes = %d, Forwarded = %d, Acked = %d, RT = %f\n",
		HeaderDTN().SequenceNumber(), TCPSequenceNo(), HeaderDTN().FragmentOffset(), HeaderDTN().BundleSize(), 
		ReceivedBytes(), Forwarded(), Acked(), RetransmitTime());
}

bool Bundle::IsInTransit()
{
	TcpDtnAgent *pTcpAgent = NULL;
	
	if (DownstreamConnection() && DownstreamConnection()->TCPAgent())
		pTcpAgent = DownstreamConnection()->TCPAgent();
	
	return pTcpAgent->BundleInTransit(*this);
}
	
/*------------------------------------------ BundleManagerDTN ---------------------------------------------------*/

BundleManagerDTN::BundleManagerDTN()
{
	pHead_ = NULL;
}

BundleManagerDTN::~BundleManagerDTN()
{
	EmptyList();
}

/* Add bundle to the queue or update bundle received. */
Bundle* BundleManagerDTN::AddBundle(Bundle newBundle) 
{
	if (!pHead_)
	{
		// printf("Adding head seqNumber = %d\n", pBundle->m_nTCPSequenceNumber);
		Bundle *pBundle = new Bundle(); *pBundle = newBundle; pBundle->InitLocalVariables();
		pHead_ = pBundle;
		// printf("Adding bundle offset %d to the list\n", pBundle->HeaderDTN().FragmentOffset());
		return pBundle;
	}

	if (newBundle.Precedes(*pHead_))
	{
		// printf("Adding first seqNumber = %d\n", pBundle->m_nTCPSequenceNumber);
		Bundle *pBundle = new Bundle(); *pBundle = newBundle; pBundle->InitLocalVariables();
		pBundle->NextBundle() = pHead_;
		pHead_ = pBundle;
		// printf("Adding bundle offset %d to the list\n", pBundle->HeaderDTN().FragmentOffset());
		return pBundle;
	}

	Bundle *pLoopBundle = pHead_;
	while (true)
	{
		Bundle *pNextBundle = pLoopBundle->NextBundle();
		
		if (newBundle.FullyContains(*pLoopBundle) && !pLoopBundle->IsFullyReceived())
		{
			/* Update existing bundle with incoming bundle data. */
			pLoopBundle->ReceivedBytes() += newBundle.HeaderDTN().Payload();
			if (pLoopBundle->ReceivedBytes() > pLoopBundle->HeaderDTN().BundleSize())
			{
				// DumpBundles();
				fprintf(stderr, "Received more bytes than the fragment size. Exiting...\n");
				exit(0);
			}
			// printf("Adding bytes to fragment offset %d. New size = %d, source addr %d.\n", pLoopBundle->HeaderDTN().FragmentOffset(), pLoopBundle->ReceivedBytes(), newBundle.HeaderDTN().SourceAddress());
			return pLoopBundle;
		}
		
		if (newBundle.FullyContains(*pLoopBundle) && (!pNextBundle || !newBundle.FullyContains(*pNextBundle)))
		{
			/*
	 	 	 * The incoming bundle must be inserted as a fragment 
	 	 	 * (the original bundle has been locally fragmented to the received bytes) 
		 	 */
	 	 	Bundle *pBundle = new Bundle(); *pBundle = newBundle; pBundle->InitLocalVariables();
			pBundle->FragmentAt(pLoopBundle->LastReceivedByte() + 1);
			// printf("Fragmented Bundle offset %d size %d sent %d orig offset %d orig size %d, source addr %d.\n", pBundle->HeaderDTN().FragmentOffset(), pBundle->HeaderDTN().BundleSize(), pBundle->Forwarded(), pBundle->OriginalOffset(), pBundle->OriginalSize(), pBundle->HeaderDTN().SourceAddress());
			pBundle->NextBundle() = pLoopBundle->NextBundle();
			pLoopBundle->NextBundle() = pBundle;
			return pBundle;
		}
		
		// if (!pNextBundle || (!pNextBundle->IsInSameADUAs(newBundle) && newBundle.Precedes(*pNextBundle)))
		if (!pNextBundle || (!newBundle.FullyContains(*pNextBundle) && newBundle.Precedes(*pNextBundle)))
		{
			/* Insert here if the next bundle is either NULL or belongs to different ADU and is ordered later than the new bundle. */
			Bundle *pBundle = new Bundle(); *pBundle = newBundle; pBundle->InitLocalVariables();
			// printf("Adding bundle offset %d to the list\n", pBundle->HeaderDTN().FragmentOffset());
			pBundle->NextBundle() = pLoopBundle->NextBundle();
			pLoopBundle->NextBundle() = pBundle;
			return pBundle;
		}

		pLoopBundle = pLoopBundle->NextBundle();
	}
	// printf("ReceivedBytes = %d, BundleSize = %d\n", pRunningBundleData->m_nReceivedBytes, pRunningBundleData->m_nBundleSize);
}

/* Completed bundles (either all segments of all fragments of a fragmented bundle or all segments of an unfragmented bundle) */
Bundle* BundleManagerDTN::GetFirstCompletedADU(Bundle* pAfterBundle)
{
	Bundle *pLoopBundle = pHead_;
	if (pAfterBundle != NULL)
		pLoopBundle = pAfterBundle->NextBundle();
	
	while (pLoopBundle)
	{
		if (pLoopBundle->IsCompleteADU())
			return pLoopBundle;
		pLoopBundle = pLoopBundle->NextBundle();
	}
	return NULL;
}

/* Gets the first unsent bundle after the pAfterBundle. */
Bundle* BundleManagerDTN::GetFirstUnSentBundle(Bundle* pAfterBundle)
{
	double fSimulatorTime = Scheduler::instance().clock();

	Bundle *pLoopBundle = pHead_;
	if (pAfterBundle != NULL)
		pLoopBundle = pAfterBundle->NextBundle();
	
	while (pLoopBundle)
	{
		double RTTime = pLoopBundle->RetransmitTime();	
		if (!pLoopBundle->Forwarded() || (RTTime > 0 && RTTime <= fSimulatorTime))
			return pLoopBundle;
		pLoopBundle = pLoopBundle->NextBundle();
	}
	return NULL;
}

/* Gets the first expired bundle after the pAfterBundle. */
Bundle* BundleManagerDTN::GetFirstExpiredBundle(Bundle* pAfterBundle)
{
	double fSimulatorTime = Scheduler::instance().clock();
	
	Bundle *pLoopBundle = pHead_;
	if (pAfterBundle != NULL)
		pLoopBundle = pAfterBundle->NextBundle();
	
	while (pLoopBundle)
	{
		if (pLoopBundle->RetransmitTime() <= fSimulatorTime)
			return pLoopBundle;
		pLoopBundle = pLoopBundle->NextBundle();
	}
	return NULL;
}

/* Find Bundle that maches the inBundle. */
Bundle* BundleManagerDTN::FindBundle(Bundle inBundle)
{
	Bundle *pLoopBundle = pHead_;

	while (pLoopBundle)
	{
		if (pLoopBundle->IsSameAs(inBundle))
			return pLoopBundle;
		pLoopBundle = pLoopBundle->NextBundle();
	}
	return NULL;
}

/* Delete all bundles that can be delivered to the application (bundles that represent a complete ADU) */
void BundleManagerDTN::Cleanup()
{
	Bundle *pLoopBundle = pHead_;
	Bundle *pPrevBundle = pHead_;
	
	while (pLoopBundle)
	{
		if (pLoopBundle->IsCompleteADU())
		{
			if (pLoopBundle == pHead_)
			{
				pHead_ = pLoopBundle->NextBundle();
				pPrevBundle = pHead_;
			}
			else
				pPrevBundle->NextBundle() = pLoopBundle->NextBundle();

			delete pLoopBundle;
			
			if (pPrevBundle)
				pLoopBundle = pPrevBundle->NextBundle();
			else
				break;
		}
		else
		{
			pPrevBundle = pLoopBundle;
			pLoopBundle = pLoopBundle->NextBundle();
		}	
	}
}

/* Remove all bundles from the list. */
void BundleManagerDTN::EmptyList()
{
	Bundle *pLoopBundle = pHead_;
		
	while (pLoopBundle)
	{
		pHead_ = pLoopBundle->NextBundle();
		delete pLoopBundle;
		pLoopBundle = pHead_;
	}
}

/*
 * Remove all instances of the pointer from the list and delete the correspoding object
 * (there actually should be at most one present, but the code accounts for multiple occurences as well).
 */
void BundleManagerDTN::RemovePointer(Bundle* pBundle)
{
	Bundle *pLoopBundle = pHead_;
	Bundle *pPrevBundle = pHead_;
	
	while (pLoopBundle)
	{
		if (pLoopBundle == pBundle)
		{
			if (pLoopBundle == pHead_)
			{
				pHead_ = pLoopBundle->NextBundle();
				pPrevBundle = pHead_;
			}
			else
				pPrevBundle->NextBundle() = pLoopBundle->NextBundle();

			if (pPrevBundle)
				pLoopBundle = pPrevBundle->NextBundle();
			else
				break;
		}
		else
		{
			pPrevBundle = pLoopBundle;
			pLoopBundle = pLoopBundle->NextBundle();
		}
	}
	
	delete pBundle;
}

/* 
 * Removes all instances of the bundle with the specified data and deletes the corresponding objects
 * (as with the above function only one should be present, even though the code accounts for multiple occurences)
 */
void BundleManagerDTN::RemoveBundle(Bundle inBundle)
{

	Bundle *pLoopBundle = pHead_;
	Bundle *pPrevBundle = pHead_;

	while (pLoopBundle)
	{
		if (inBundle.IsSameAs(*pLoopBundle))
		{
			if (pLoopBundle == pHead_)
			{
				pHead_ = pLoopBundle->NextBundle();
				pPrevBundle = pHead_;
			}
			else
			{
				pPrevBundle->NextBundle() = pLoopBundle->NextBundle();
			}

			delete pLoopBundle;
			
			if (pPrevBundle)
				pLoopBundle = pPrevBundle->NextBundle();
			else
				break;
		}
		else
		{
			pPrevBundle = pLoopBundle;
			pLoopBundle = pLoopBundle->NextBundle();
		}
	}
}

/* 
 * Removes all fragments that constitute the original bundle pBundle belongs to.
 * It can be a single bundle in case fragmentation did not take place on this node.
 */
bool BundleManagerDTN::CleanIfOriginalAcked(Bundle inBundle)
{
	/* 
	 * If the original bundle has not been completely received we cannot remove fragments even though custody for them may
	 * have been successfully transferred. Otherwise we will not know which part of the original bundle incoming bytes 
	 * of the same bundle belong to.
	 */
	if (!OriginalBundleAcked(inBundle))
		return false;
		
	// printf("Original Bundle Completed offset %d size %d sent %d.\n", inBundle.HeaderDTN().FragmentOffset(), inBundle.HeaderDTN().BundleSize(), inBundle.Forwarded());
	
	/* Remove all bundles belonging to the original bundle of the pBundle. */
	Bundle* pLoopBundle = pHead_;
	while (pLoopBundle)
	{
		Bundle *pNextBundle = pLoopBundle->NextBundle();
		if (pLoopBundle->OriginalFullyContains(inBundle))
			RemoveBundle(*pLoopBundle);
		pLoopBundle = pNextBundle;
	}
	return true;
}

/* 
 * Checks whether all fragments of the original bundle that pBundle belongs to have been fully received.
 * It does so by checking that the sum of the received bytes of the fragments equals the size of the original bundle.
 */
bool BundleManagerDTN::OriginalBundleAcked(Bundle inBundle)
{
	int nOrigBundleBytes = 0;
	int nOrigBundleSize = -1;
	
	Bundle *pLoopBundle = pHead_;
	
	while (pLoopBundle)
	{
		if (pLoopBundle->OriginalFullyContains(inBundle) && pLoopBundle->Acked())
		{
			nOrigBundleBytes += pLoopBundle->ReceivedBytes();
			nOrigBundleSize = pLoopBundle->OriginalSize();
		}
		
		if (nOrigBundleBytes == nOrigBundleSize)
			return true;
		else if (inBundle.Precedes(*pLoopBundle))
			return false;

		pLoopBundle = pLoopBundle->NextBundle();
	}
	return false;
}

/* 
 * Checks whether all fragments of the original bundle that pBundle belongs to have been fully received.
 * It does so by checking that the sum of the received bytes of the fragments equals the size of the original bundle.
 */
bool BundleManagerDTN::OriginalBundleCompleted(Bundle inBundle)
{
	int nOrigBundleBytes = 0;
	int nOrigBundleSize = -1;
	
	Bundle *pLoopBundle = pHead_;
	
	//printf("Bundle: offset %d, size %d, source addr %d, timestamp %f.\n", pBundle->HeaderDTN().FragmentOffset(),  pBundle->HeaderDTN().BundleSize(), pBundle->HeaderDTN().SourceAddress(), pBundle->HeaderDTN().Timestamp());

	while (pLoopBundle)
	{
		if (pLoopBundle->OriginalFullyContains(inBundle))
		{
			if (pLoopBundle->IsFullyReceived())
				nOrigBundleBytes += pLoopBundle->HeaderDTN().BundleSize();
			else
				nOrigBundleBytes += pLoopBundle->ReceivedBytes();
			nOrigBundleSize = pLoopBundle->OriginalSize();
		}
		// printf("Loop: Orig offset %d orig size %d, offset %d, size %d, received %d, source addr %d, timestamp %f.\n", pLoopBundle->OriginalOffset(), pLoopBundle->OriginalSize(), pLoopBundle->HeaderDTN().FragmentOffset(),  pLoopBundle->HeaderDTN().BundleSize(), pLoopBundle->ReceivedBytes(), pLoopBundle->HeaderDTN().SourceAddress(), pLoopBundle->HeaderDTN().Timestamp());
		
		// printf("origBundleBytes = %d, origBundleSize = %d\n", nOrigBundleBytes,nOrigBundleSize);
		
		if (nOrigBundleBytes == nOrigBundleSize)
			return true;
		else if (!inBundle.IsInSameADUAs(*pLoopBundle) && inBundle.Precedes(*pLoopBundle))
			return false;

		pLoopBundle = pLoopBundle->NextBundle();
	}
	return false;
}

/* Returns the number of bundles stored in the manager. */
int BundleManagerDTN::BundlesStored()
{
	int nBundles = 0;

	Bundle *pLoopBundle = pHead_;

	while (pLoopBundle)
	{
		nBundles++;
		pLoopBundle = pLoopBundle->NextBundle();
	}

	return nBundles;
}

/* Returns the bytes stored in the manager. */
int BundleManagerDTN::MemoryAllocated()
{
	int nTotalSize = 0;

	Bundle *pLoopBundle = pHead_;

	while (pLoopBundle)
	{
		nTotalSize += pLoopBundle->HeaderDTN().BundleSize();
		pLoopBundle = pLoopBundle->NextBundle();
	}

	return nTotalSize;
}

/* Merges all fragments in the bundle list. */
void BundleManagerDTN::MergeAllFragments()
{
	Bundle *pLoopBundle = pHead_;
		
	while (pLoopBundle)
	{
		if (pLoopBundle->NextBundle() && pLoopBundle->CanBeMergedWith(*(pLoopBundle->NextBundle())))
		{
			Bundle *pTempBundle = pLoopBundle->NextBundle()->NextBundle();
			pLoopBundle->MergeWith(*(pLoopBundle->NextBundle()));
			delete pLoopBundle->NextBundle();
			pLoopBundle->NextBundle() = pTempBundle;
		}
		else
		{
			pLoopBundle = pLoopBundle->NextBundle();
		}
	}
}

/* Fragments all bundles that are partially received for immediate transmission. */
void BundleManagerDTN::FragmentAllBundles()
{
	Bundle *pLoopBundle = pHead_;
		
	while (pLoopBundle)
	{
		pLoopBundle->FragmentReceivedBytes();
		pLoopBundle = pLoopBundle->NextBundle();
	}
}

void BundleManagerDTN::DumpBundles()
{
	int nBundles = 0;

	Bundle *pLoopBundle = pHead_;

	while (pLoopBundle)
	{
		nBundles++;
		printf("Bundle %d: ", nBundles);
		pLoopBundle->Dump();
		pLoopBundle = pLoopBundle->NextBundle();
	}
}

Bundle* BundleManagerDTN::FindOriginalBundle(Bundle inBundle)
{
	Bundle *pLoopBundle = pHead_;
	
	while (pLoopBundle)
	{
		if (pLoopBundle->OriginalFullyContains(inBundle))
			return pLoopBundle;

		pLoopBundle = pLoopBundle->NextBundle();
	}
	return NULL;
}

double BundleManagerDTN::FindEarlierRTTime()
{
	double minRTTime = 0;
	Bundle *pLoopBundle = pHead_;
	
	while (pLoopBundle)
	{
		double RTTime = pLoopBundle->RetransmitTime();
		if ((RTTime > 0) && (RTTime < minRTTime || minRTTime == 0) )
			minRTTime = RTTime;

		pLoopBundle = pLoopBundle->NextBundle();
	}
		
	return minRTTime;
}

// ----------------------------------------------- Dtn Timers -----------------------------------------------------//

void DtnAgentRVTimer::expire(Event*)
{  
	t_->RVTimerHandler(); 
}

void DtnAgentRTTimer::expire(Event*)
{ 
	t_->RTTimerHandler(); 
}

double DtnAgentRTTimer::SetoffTime()
{
	if (status()  == TIMER_PENDING)
		return event_.time_;
	else
		return 0;
}

// ----------------------------------------------- DtnConnection ------------------------------------------------ //

DtnConnection::DtnConnection(int nConnectionID, DtnAgent *pOwnerDtnAgent, DtnAgent *pPeerDtnAgent)
{
	nConnectionID_		= nConnectionID;
	pOwnerDtnAgent_		= pOwnerDtnAgent;
	pPeerDtnAgent_		= pPeerDtnAgent;
	pTCPAgent_		= NULL;
	pTCPSink_		= NULL;
	pNextConnection_	= NULL;
}

DtnConnection::~DtnConnection()
{
	if (pTCPAgent_)
		delete pTCPAgent_;
	if (pTCPSink_)
		delete pTCPSink_;
}

void DtnConnection::Connect()
{
	if (pTCPAgent_ || pTCPSink_)
		return;
	Tcl& tcl = Tcl::instance();
	tcl.evalf("[Simulator instance] connect-tcp %s %s", pOwnerDtnAgent_->name(), pPeerDtnAgent_->name());
}

// ----------------------------------------------- ConnectionManager ---------------------------------------------- //

ConnectionManager::ConnectionManager(DtnAgent *pOwnerDtnAgent)
{
	pOwnerDtnAgent_			= pOwnerDtnAgent;
	pHeadDownstream_		= NULL;
	pHeadUpstream_			= NULL;
	bDownstreamConnectionsSet_	= false;
}

ConnectionManager::~ConnectionManager()
{
	EmptyList();
}

/* Remove all connections from the list. */
void ConnectionManager::EmptyList()
{
	DtnConnection *pLoopConnection = pHeadDownstream_;
	while (pLoopConnection)
	{
		pHeadDownstream_ = pLoopConnection->NextConnection();
		delete pLoopConnection;
		pLoopConnection = pHeadDownstream_;
	}
	
	pLoopConnection = pHeadUpstream_;
	while (pLoopConnection)
	{
		pHeadUpstream_ = pLoopConnection->NextConnection();
		delete pLoopConnection;
		pLoopConnection = pHeadUpstream_;
	}
}
	
int ConnectionManager::DownstreamConnections()
{
	int nConnections = 0;
	
	DtnConnection *pLoopConnection = pHeadDownstream_;

	while (pLoopConnection)
	{
		pLoopConnection = pLoopConnection->NextConnection();
		nConnections++;
	}
	return nConnections;
}

int ConnectionManager::UpstreamConnections()
{
	int nConnections = 0;
	
	DtnConnection *pLoopConnection = pHeadUpstream_;

	while (pLoopConnection)
	{
		pLoopConnection = pLoopConnection->NextConnection();
		nConnections++;
	}
	return nConnections;
}

void ConnectionManager::SetupDownstreamConnections()
{
	if (bDownstreamConnectionsSet_)
		return;
		
	DtnConnection *pLoopConnection = pHeadDownstream_;

	while (pLoopConnection)
	{
		pLoopConnection->Connect();
		pLoopConnection = pLoopConnection->NextConnection();
	}
	
	bDownstreamConnectionsSet_ = true;
}

void ConnectionManager::AddDownstreamConnection(DtnAgent *pPeerDtnAgent)
{
	int nConnectionID = 0;
	
	DtnConnection *pLoopConnection = pHeadDownstream_;
	DtnConnection *pPreviousConnection = pHeadDownstream_;

	while (pLoopConnection)
	{
		pPreviousConnection = pLoopConnection;
		pLoopConnection = pLoopConnection->NextConnection();
		nConnectionID++;
	}
	DtnConnection *pNewConnection 	= new DtnConnection(nConnectionID, pOwnerDtnAgent_, pPeerDtnAgent);
	
	if (pPreviousConnection == NULL)
		pHeadDownstream_ = pNewConnection;
	else
	{
		pNewConnection->NextConnection() = pPreviousConnection->NextConnection();
		pPreviousConnection->NextConnection() = pNewConnection;
	}
}

void ConnectionManager::AddUpstreamConnection(DtnAgent *pPeerDtnAgent)
{
	int nConnectionID = 0;

	DtnConnection *pLoopConnection = pHeadUpstream_;
	DtnConnection *pPreviousConnection = pHeadUpstream_;

	while (pLoopConnection)
	{
		pPreviousConnection = pLoopConnection;
		pLoopConnection = pLoopConnection->NextConnection();
		nConnectionID++;
	}
	
	DtnConnection *pNewConnection 	= new DtnConnection(nConnectionID, pOwnerDtnAgent_, pPeerDtnAgent);
	
	if (pPreviousConnection == NULL)
		pHeadUpstream_ = pNewConnection;
	else
	{
		pNewConnection->NextConnection() = pPreviousConnection->NextConnection();
		pPreviousConnection->NextConnection() = pNewConnection;
	}
}

bool ConnectionManager::SetDownstreamTCP(DtnAgent *pDtnAgent, TcpDtnAgent *pTcpAgent, TcpDtnSink *pTcpSink)
{
	DtnConnection *pLoopConnection = pHeadDownstream_;

	while (pLoopConnection)
	{
		if (pLoopConnection->PeerDtnAgent() == pDtnAgent)
		{
			pTcpAgent->SetDTNAgent(pOwnerDtnAgent_);
			pTcpSink->SetDTNAgent(pOwnerDtnAgent_);
			pLoopConnection->TCPAgent() = pTcpAgent;
			pLoopConnection->TCPSink()	= pTcpSink;
			return true;
		}

		pLoopConnection = pLoopConnection->NextConnection();
	}
	return false;
}

bool ConnectionManager::SetUpstreamTCP(DtnAgent *pDtnAgent, TcpDtnAgent *pTcpAgent, TcpDtnSink *pTcpSink)
{
	DtnConnection *pLoopConnection = pHeadUpstream_;

	while (pLoopConnection)
	{
		if (pLoopConnection->PeerDtnAgent() == pDtnAgent)
		{
			pTcpAgent->SetDTNAgent(pOwnerDtnAgent_);
			pTcpSink->SetDTNAgent(pOwnerDtnAgent_);		
			pLoopConnection->TCPAgent() = pTcpAgent;
			pLoopConnection->TCPSink()	= pTcpSink;
			return true;
		}

		pLoopConnection = pLoopConnection->NextConnection();
	}
	return false;
}
	
DtnConnection* ConnectionManager::GetDownstreamConnection(int nConnectionID)
{
	DtnConnection *pLoopConnection = pHeadDownstream_;

	while (pLoopConnection)
	{
		if (pLoopConnection->ConnectionID() == nConnectionID)
			return pLoopConnection;

		pLoopConnection = pLoopConnection->NextConnection();
	}
	return NULL;
}

DtnConnection* ConnectionManager::GetUpstreamConnection(int nConnectionID)
{
	DtnConnection *pLoopConnection = pHeadUpstream_;

	while (pLoopConnection)
	{
		if (pLoopConnection->ConnectionID() == nConnectionID)
			return pLoopConnection;

		pLoopConnection = pLoopConnection->NextConnection();
	}
	return NULL;
}
/*
DtnConnection* ConnectionManager::GetConnection(DtnAgent *pPeerDtnAgent)
{
	DtnConnection *pLoopConnection = pHeadDownstream_;

	while (pLoopConnection)
	{
		if (pLoopConnection->PeerDtnAgent() == pPeerDtnAgent)
			return pLoopConnection;

		pLoopConnection = pLoopConnection->NextConnection();
	}
	
	pLoopConnection = pHeadUpstream_;

	while (pLoopConnection)
	{
		if (pLoopConnection->PeerDtnAgent() == pPeerDtnAgent)
			return pLoopConnection;

		pLoopConnection = pLoopConnection->NextConnection();
	}
	return NULL;
}

DtnConnection* ConnectionManager::GetConnection(TcpDtnAgent *pTCPAgent)
{
	DtnConnection *pLoopConnection = pHeadDownstream_;

	while (pLoopConnection)
	{
		if (pLoopConnection->TCPAgent() == pTCPAgent)
			return pLoopConnection;

		pLoopConnection = pLoopConnection->NextConnection();
	}
	
	pLoopConnection = pHeadUpstream_;

	while (pLoopConnection)
	{
		if (pLoopConnection->TCPAgent() == pTCPAgent)
			return pLoopConnection;

		pLoopConnection = pLoopConnection->NextConnection();
	}
	return NULL;
}
*/
DtnConnection* ConnectionManager::GetConnection(TcpDtnSink *pTCPSink)
{
	DtnConnection *pLoopConnection = pHeadDownstream_;
	while (pLoopConnection)
	{
		if (pLoopConnection->TCPSink() == pTCPSink)
		{
			return pLoopConnection;
		}
		pLoopConnection = pLoopConnection->NextConnection();
	}
	pLoopConnection = pHeadUpstream_;

	while (pLoopConnection)
	{
		if (pLoopConnection->TCPSink() == pTCPSink)
			return pLoopConnection;

		pLoopConnection = pLoopConnection->NextConnection();
	}
	return NULL;
}

// ----------------------------------------------- DtnAgent ------------------------------------------------------ //

static class DtnClass : public TclClass {
public:
	DtnClass() : TclClass("Agent/DTN") {}
	TclObject* create(int, const char*const*) {
		return (new DtnAgent());
	}
} class_dtn;

DtnAgent::DtnAgent() : Agent(PT_DTN), connectionManager_(this), RVtimer_(this), RTtimer_(this)
{
	m_nNextBundleSeq	= 1;
	m_nExpectedSegmentSeq	= 1;
	m_nExpectedBundleSeq	= 1;

	fLastFlushTime_		= 0;
	fLastUtilization_	= 0;
	nIsBSNode_		= 0;
	m_nBytesReceived	= 0;

	nLastConnectionIndex_	= -1;
	
	// pMac_		= NULL;
	
	bind("nBDProduct_", &nBDProduct_);
	bind("nTransferSize_", &nTransferSize_);
	bind("nIsBSNode_", &nIsBSNode_);	
	bind("nMaxBundleSize_", &nMaxBundleSize_);
	bind("fTimeout_", &fTimeout_);
	bind("nStorageSpace_", &nStorageSpace_);
	bind("nTCPPacketSize_", &nTCPPacketSize_);
}


DtnAgent::~DtnAgent()
{

}

int DtnAgent::command(int argc, const char*const* argv)
{
	if (argc == 2) {
		;
	}
	else if (argc == 3) {
		if (strcmp(argv[1], "add-downstream-agent") == 0) {
			if (AddDtnAgent(argv[2], "downstream"))
				return (TCL_OK);
			else
				return (TCL_ERROR);
		}
		if (strcmp(argv[1], "add-upstream-agent") == 0 ) {
			if (AddDtnAgent(argv[2], "upstream"))
				return (TCL_OK);
			else
				return (TCL_ERROR);
		}
	}
	else if (argc == 5) {
		if (strcmp(argv[1], "attach-down-tcp") == 0) {
			if (AttachTcp(argv[2], argv[3], argv[4], "downstream"))
				return (TCL_OK);
			else
				return (TCL_ERROR);
		}
		if (strcmp(argv[1], "attach-up-tcp") == 0) {
			if (AttachTcp(argv[2], argv[3], argv[4], "upstream"))
				return (TCL_OK);
			else
				return (TCL_ERROR);
		}
	}
	return (Agent::command(argc, argv));
}

bool DtnAgent::AttachTcp(const char* DtnAgentHandle, const char* TcpAgentHandle, const char* TcpSinkHandle, const char* agentType)
{
	TclObject* pDtnAgent = Tcl::instance().lookup(DtnAgentHandle);
	TclObject* pTcpAgent = Tcl::instance().lookup(TcpAgentHandle);
	TclObject* pTcpSink = Tcl::instance().lookup(TcpSinkHandle);

	if (pDtnAgent == NULL)
	{
		fprintf(stderr, "Bad object name %s\n", DtnAgentHandle);
		return false;
	}
	if (pTcpAgent == NULL)
	{
		fprintf(stderr, "Bad object name %s\n", TcpAgentHandle);
		return false;
	}
	if (pTcpSink == NULL)
	{
		fprintf(stderr, "Bad object name %s\n", TcpSinkHandle);
		return false;
	}
	if (strcmp(agentType, "downstream") == 0)
	{
		connectionManager_.SetDownstreamTCP((DtnAgent*) pDtnAgent, (TcpDtnAgent*) pTcpAgent, (TcpDtnSink*) pTcpSink);
		return true;
	}
	if (strcmp(agentType, "upstream") == 0)
	{
		connectionManager_.SetUpstreamTCP((DtnAgent*) pDtnAgent, (TcpDtnAgent*) pTcpAgent, (TcpDtnSink*) pTcpSink);
		return true;
	}
	
	return false;
}

bool DtnAgent::AddDtnAgent(const char* agentHandle, const char* agentType)
{
	TclObject* pObject = Tcl::instance().lookup(agentHandle);
	if (pObject == NULL)
	{
		fprintf(stderr, "Bad object name %s\n", agentHandle);
		return false;
	}
	if (strcmp(agentType, "downstream") == 0)
	{
		connectionManager_.AddDownstreamConnection((DtnAgent *) pObject);
		return true;
	}
	if (strcmp(agentType, "upstream") == 0)
	{
		connectionManager_.AddUpstreamConnection((DtnAgent *) pObject);
		return true;
	}
	
	return false;
}

void DtnAgent::sendmsg(int nbytes, const char* /*flags*/)
{
	connectionManager_.SetupDownstreamConnections();
	
	int nBundles		= ceil((double) nbytes / nMaxBundleSize_);
	int nLeftoverBytes	= nbytes; // DTN header size

	for (int i = m_nNextBundleSeq; i < m_nNextBundleSeq + nBundles; i++)
	{
		Bundle newBundle;
		newBundle.HeaderDTN().SequenceNumber()		= i;
		newBundle.HeaderDTN().SourceAddress()		= addr();
		newBundle.HeaderDTN().Timestamp()		= Scheduler::instance().clock();
		newBundle.HeaderDTN().FragmentOffset()		= (i - m_nNextBundleSeq) * nMaxBundleSize_;
		if (nLeftoverBytes > nMaxBundleSize_)
			newBundle.HeaderDTN().BundleSize()	= nMaxBundleSize_;
		else
			newBundle.HeaderDTN().BundleSize()	= nLeftoverBytes;
		newBundle.HeaderDTN().ADUSize()			= nbytes;
		newBundle.HeaderDTN().Payload()			= newBundle.HeaderDTN().BundleSize();
		newBundle.HeaderDTN().SegmentNumber()		= 0;
		newBundle.HeaderDTN().IsStatusReport()		= 0;
		newBundle.HeaderDTN().CustodyAccepted()		= 0;
		newBundle.HeaderDTN().LeadingBytes()		= 1;
		newBundle.HeaderDTN().ADUSize()			= nbytes;
		newBundle.ReceivedBytes()			= newBundle.HeaderDTN().BundleSize();
		newBundle.RetransmitTime() 			= 0;
		
		debug("DTN (%d): Creating bundle %d, fragment %d, of size %d, recBytes %d, retrans %f!\n", addr(), newBundle.HeaderDTN().SequenceNumber(), 
			newBundle.HeaderDTN().FragmentOffset(), newBundle.HeaderDTN().BundleSize(), newBundle.ReceivedBytes(), newBundle.RetransmitTime());
		Bundle *pStoredBundle = bundleManagerDTN_.AddBundle(newBundle);
		SendDataBundle(pStoredBundle, pStoredBundle->HeaderDTN().BundleSize() + DTN_HEADER_SIZE);
		nLeftoverBytes = nLeftoverBytes - pStoredBundle->HeaderDTN().BundleSize();
	}
	m_nNextBundleSeq++;
}

void DtnAgent::ReceiveBundlePayload(Bundle inBundle, TcpDtnSink *pTcpDtnSink)
{
	inBundle.UpstreamConnection() = connectionManager_.GetConnection(pTcpDtnSink);
	
	if (!inBundle.UpstreamConnection())
	{
		debug("DTN (%d): Incoming connection not found for bundle = %d, fragment = %d, size = %d\n", addr(), 
			inBundle.HeaderDTN().SequenceNumber(), inBundle.HeaderDTN().FragmentOffset(), inBundle.HeaderDTN().BundleSize());
		return;
	}
	
	if (inBundle.HeaderDTN().Payload() == 0)
	{
		/* Empty DTN payload bundle received. Just a safeguard since the TcpDtnSink does not allow such a bundle to reach the DtnAgent. */
		debug("DTN (%d): Rejecting empty bundle = %d, fragment = %d, size = %d, payload = %d\n", addr(), 
			inBundle.HeaderDTN().SequenceNumber(), inBundle.HeaderDTN().FragmentOffset(), inBundle.HeaderDTN().BundleSize(),
			inBundle.HeaderDTN().Payload());
		return;
	}

	if (ProcessStatusReport(inBundle))
		return;

	if (RejectBundle(inBundle))
		return;

	if (inBundle.HeaderDTN().LeadingBytes())
	{
		/* The received bundle segment contains the leading packets of the bundle. Decrease payload by the header size. */
		inBundle.HeaderDTN().Payload() -= DTN_HEADER_SIZE;
	}

	debug("DTN (%d): Receiving bundle = %d, fragment = %d, bytes = %d, size = %d, lead = %d\n", addr(), inBundle.HeaderDTN().SequenceNumber(),  
		inBundle.HeaderDTN().FragmentOffset(), inBundle.HeaderDTN().Payload(), inBundle.HeaderDTN().BundleSize(), inBundle.HeaderDTN().LeadingBytes());	
	
	/* At his point inBundle is a data bundle custody for which has not been rejected, so it is inserted into the bundle manager. */
	Bundle *pStoredBundle = bundleManagerDTN_.AddBundle(inBundle);
		
	if (bundleManagerDTN_.OriginalBundleCompleted(inBundle))
	{
		/* Check if this bundle completes the original bundle (before possible local fragmentation) and send custody acceptance report. */
		SendCustodyReport(inBundle, true);
	}
	
	if (app_)
	{
		/* Merge all fragments that can be merged and deliver fully received bundles to the application. */
		bundleManagerDTN_.MergeAllFragments();
		Bundle *pRetrievedBundle = bundleManagerDTN_.GetFirstCompletedADU(NULL);
		while (pRetrievedBundle)
		{
			app_->recv(pRetrievedBundle->HeaderDTN().BundleSize());
			pRetrievedBundle = bundleManagerDTN_.GetFirstCompletedADU(pRetrievedBundle);
		}
		bundleManagerDTN_.Cleanup();
	}
	// else if (m_pDownstreamSend)
	else if (connectionManager_.DownstreamConnections() > 0)
	{
		connectionManager_.SetupDownstreamConnections();
		/* A downstream agent exists so the bundle must be forwarded. */
		if (!nIsBSNode_)
		{
			/* This is a relaying node (not the base station that may employ the Rendezvous mechanism so route the packet normally. */
			if (!pStoredBundle->Suspended())
			{
				/* Forward bundle unless routing has been suspended for it. */
				pStoredBundle->Forwarded() = 1;
				pStoredBundle->HeaderDTN().LeadingBytes() = inBundle.HeaderDTN().LeadingBytes();
				SendDataBundle(pStoredBundle, inBundle.HeaderDTN().Payload() + inBundle.HeaderDTN().LeadingBytes() * DTN_HEADER_SIZE);
				// printf("DTN (%d): Cut-through bundle = %d, fragment = %d, bytes = %d, size = %d\n", addr(), inBundle.HeaderDTN().SequenceNumber(),  
				//	inBundle.HeaderDTN().FragmentOffset(), inBundle.HeaderDTN().Payload(), inBundle.HeaderDTN().BundleSize());	

				// Add option for non cut-through routing.
			}
		}
		else
		{
			/* Increment received bytes for the RV mechanism and set the initial timer if idle. */
			// printf("File transfer of %d (tr size = %d, bytes received = %d)!\n", nTotalBytesReceived_, nTransferSize_, nbytes - 40);
			// printf("Bytes received = %d!\n", receivedBundleData.m_nPayload);
			m_nBytesReceived += inBundle.HeaderDTN().Payload();
			if ( RVtimer_.status() == TIMER_IDLE )
			{
				double fSimulatorTime = Scheduler::instance().clock();
				double fReceptionInterval = fSimulatorTime - fLastFlushTime_;
				double fNextTimeout;
			
				if (fReceptionInterval > 0.5)
					fNextTimeout = 0.5;
				else
					fNextTimeout = fReceptionInterval;
							
				// Tcl& tcl = Tcl::instance();
				// tcl.evalf("[Simulator instance] puts-ns-traceall \"Buffer emptied %f (initial) at %f. Next timeout in %f.\"",
					// m_bundleDataManager.MemoryAllocated()/1000.0, Scheduler::instance().clock(), fNextTimeout);
				// printf("Base Station (initial): Time = %f, Buffer emptied %d, Total bytes = %d, Channel utilization = %f, (old = %f)!\n", 
					// fSimulatorTime, nBytesReceived_,  nTotalBytesReceived_, pMac_->GetUtilization(), pMac_->GetUtilizationOld());
				SendOutstandingBundles();
				debug("First timeout at %f!\n", fSimulatorTime + fNextTimeout);
				RVtimer_.resched(fNextTimeout);

				fLastFlushTime_ 	= fSimulatorTime;
				m_nBytesReceived 	= 0;
			}
		}
	}
}

bool DtnAgent::ProcessStatusReport(Bundle inBundle)
{
	if (!inBundle.HeaderDTN().IsStatusReport())
		return false;
		
	Bundle *pStoredBundle = bundleManagerDTN_.FindBundle(inBundle);
	
	if (inBundle.HeaderDTN().CustodyAccepted())
	{
		/* Process custody acceptance report. */

		/* Mark stored bundle as acknowledged. */
		if (pStoredBundle)
		{
			pStoredBundle->Acked() = 1;
			/* Cancel possible pending transmission for the acknowledged bundle. */
			if (pStoredBundle->DownstreamConnection() && pStoredBundle->DownstreamConnection()->TCPAgent())
				pStoredBundle->DownstreamConnection()->TCPAgent()->CancelBundleTransmission(inBundle);
		}

		/* Clean storage only if the whole of the original bundle (in case it was fragmented at this node) has been acknowledged. */
		bundleManagerDTN_.CleanIfOriginalAcked(inBundle);
		
		debug("DTN (%d): Custody acceptance received for bundle = %d, fragment = %d, size = %d. Bundles stored = %d\n", addr(), 
			inBundle.HeaderDTN().SequenceNumber(), inBundle.HeaderDTN().FragmentOffset(), inBundle.HeaderDTN().BundleSize(),
			bundleManagerDTN_.BundlesStored());
			
		/* Update timer in case the timer was set for the acknowledged bundle. */
		UpdateRTTimer();
	}
	else if (inBundle.HeaderDTN().CustodyRejected())
	{
		/* Process custody rejection report. */
		debug("DTN (%d): Custody rejection received for bundle = %d, fragment = %d, size = %d. Bundles stored = %d\n", addr(), 
			inBundle.HeaderDTN().SequenceNumber(), inBundle.HeaderDTN().FragmentOffset(), inBundle.HeaderDTN().BundleSize(),
			bundleManagerDTN_.BundlesStored());
		
		if (pStoredBundle)
		{
			/* Mark bundle as not forwarded and suspend cut-through routing. */
			pStoredBundle->Forwarded() = 0;
			pStoredBundle->Suspended() = 1;
			/* Cancel possible pending transmission for the rejected bundle. */
			if (pStoredBundle->DownstreamConnection() && pStoredBundle->DownstreamConnection()->TCPAgent())
				pStoredBundle->DownstreamConnection()->TCPAgent()->CancelBundleTransmission(inBundle);
			
			if (fTimeout_ > 0) 
			{ 
				/* Set retransmission time for the bundle. */ 
				SetBundleRTTime(pStoredBundle); 
 			} 
 			else 
			{ 
 				bundleManagerDTN_.RemoveBundle(inBundle); 
 				double fOccupancy = (1.0 * bundleManagerDTN_.MemoryAllocated() ) / nStorageSpace_; 
				printf("Removing bundle at node %d, New BufferOccupancy: %f\n", addr() / 2048, fOccupancy); 
			} 				
		}
	}
	else
		return false;
	
	return true;
}

bool DtnAgent::RejectBundle(Bundle inBundle)
{	
	if (!inBundle.HeaderDTN().LeadingBytes() && !bundleManagerDTN_.FindOriginalBundle(inBundle))
	{
		/* Bytes cannot be accepted as the bundle header is missing. */
		debug("DTN (%d): Rejecting bytes for bundle = %d, fragment = %d, size = %d. \n", addr(), 
			inBundle.HeaderDTN().SequenceNumber(), inBundle.HeaderDTN().FragmentOffset(), inBundle.HeaderDTN().BundleSize());
		return true;
	}
	
	// if (pStoredBundle && bundleManagerDTN_.OriginalBundleCompleted(*pStoredBundle)) // CHANGE !!!
	if (bundleManagerDTN_.OriginalBundleCompleted(inBundle))
	{
		/* Duplicate bundle received. */
		debug("DTN (%d): Rejecting duplicate bundle = %d, fragment = %d, size = %d. Bundles stored = %d\n", addr(), 
			inBundle.HeaderDTN().SequenceNumber(), inBundle.HeaderDTN().FragmentOffset(), inBundle.HeaderDTN().BundleSize(),
			bundleManagerDTN_.BundlesStored());
		return true;
	}
	
	/* 
	* Receiving leading bytes of a partially received bundle. Remove bundle from manager.
	* Care must be taken in case the bundle has been fragmented and forwarded to the next hop. 
	*/
	Bundle *pStoredBundle = bundleManagerDTN_.FindBundle(inBundle);
	
	if (inBundle.HeaderDTN().LeadingBytes() && pStoredBundle)
		bundleManagerDTN_.RemoveBundle(inBundle);
		
	if (inBundle.HeaderDTN().LeadingBytes() && nStorageSpace_ > 0 &&
		(bundleManagerDTN_.MemoryAllocated() + inBundle.HeaderDTN().BundleSize() > nStorageSpace_) && !app_)
	{
		/* Accepting the incoming bundle will exceed available storage space. */
		SendCustodyReport(inBundle, false);
		return true;
	}

	return false;
}

void DtnAgent::SendDataBundle(Bundle *pBundle, int nBytes)
{
	DtnConnection *pDownstreamConnection = pBundle->DownstreamConnection();
	
	if (!pDownstreamConnection)
	{
		if (connectionManager_.DownstreamConnections() == 0)
		{
			debug("No Downstream Connections!!!\n");
			return;
		}
		// Add routing here
		/* For now round robin forward to all outgoing connections. */
		nLastConnectionIndex_ = (nLastConnectionIndex_ + 1) % connectionManager_.DownstreamConnections();
		pDownstreamConnection = connectionManager_.GetDownstreamConnection(nLastConnectionIndex_);
		pBundle->DownstreamConnection() = pDownstreamConnection;
	}

	pDownstreamConnection->Connect();
	pDownstreamConnection->TCPAgent()->sendBundle(*pBundle, nBytes);
}

void DtnAgent::SendCustodyReport(Bundle inBundle, bool bAccept)
{
	DtnConnection *pUpstreamConnection = inBundle.UpstreamConnection();
	if (!pUpstreamConnection)
	{
		debug("No Upstream connection for received bundle!!!\n");
		return;
	}

	pUpstreamConnection->Connect();
	
	Bundle reportBundle;
	reportBundle = inBundle;
	
	if (bAccept)
	{
		debug("DTN (%d): Sending custody report for bundle = %d, fragment = %d, size = %d\n", addr(), inBundle.HeaderDTN().SequenceNumber(),
			inBundle.HeaderDTN().FragmentOffset(), inBundle.HeaderDTN().BundleSize());
		reportBundle.AcceptCustody();
	}
	else
	{
		debug("DTN (%d): Sending custody rejection for bundle = %d, fragment = %d, size = %d\n", addr(), inBundle.HeaderDTN().SequenceNumber(),
			inBundle.HeaderDTN().FragmentOffset(), inBundle.HeaderDTN().BundleSize());
		reportBundle.RejectCustody();
	}
	
	pUpstreamConnection->TCPAgent()->sendBundle(reportBundle, DTN_HEADER_SIZE);
}

/* Callback from the TcpDtn agent when the last part of a bundle has been sent so that the RT timer is set. */
void DtnAgent::BundleTransmitted(Bundle inBundle)
{
	if (fTimeout_ == 0 || inBundle.HeaderDTN().IsStatusReport() || nIsBSNode_)
		return;

	Bundle *pStoredBundle = bundleManagerDTN_.FindBundle(inBundle);
	/* Do not set retransmission for partially received bundles that are cut-through routed. */
	if (pStoredBundle && pStoredBundle->IsFullyReceived())
		SetBundleRTTime(pStoredBundle);
}

/* Set RT time for the specified bundle. */
void DtnAgent::SetBundleRTTime(Bundle *pBundle)
{
	double fSimulatorTime = Scheduler::instance().clock();
	double fNextRTInterval = fTimeout_;
	pBundle->RetransmitTime() = fSimulatorTime + fNextRTInterval;
	debug("DTN (%d): Setting retransmit at %f for bundle = %d, fragment = %d, size = %d\n", addr(), pBundle->RetransmitTime(),
		pBundle->HeaderDTN().SequenceNumber(), pBundle->HeaderDTN().FragmentOffset(), pBundle->HeaderDTN().BundleSize());
	if (RTtimer_.status() == TIMER_IDLE || RTtimer_.SetoffTime() > pBundle->RetransmitTime())
	{
		RTtimer_.resched(fNextRTInterval);
		debug("DTN (%d): Rescheduling timer!\n", addr());
	}
}

/* Set timer to the earliest RTTime in the bundle manager. */
void DtnAgent::UpdateRTTimer()
{
	if (fTimeout_ == 0 || nIsBSNode_ || bundleManagerDTN_.FindEarlierRTTime() == 0)
		return;
		
	double fSimulatorTime = Scheduler::instance().clock();
	RTtimer_.resched(bundleManagerDTN_.FindEarlierRTTime() - fSimulatorTime);
}

/* Called when the RTTimer expires. Retransmits all bundles and resets the timer. */
void DtnAgent::RTTimerHandler()
{
	if (fTimeout_ == 0 || nIsBSNode_)
		return;

	/* Retransmit all bundles for which the RTTime value has passed. */
	Bundle *pExpiredBundle = bundleManagerDTN_.GetFirstExpiredBundle(NULL);
	while (pExpiredBundle)
	{
		if (!pExpiredBundle->IsInTransit())
		{
			pExpiredBundle->HeaderDTN().LeadingBytes() = 1;
			pExpiredBundle->RetransmitTime() = 0;
			pExpiredBundle->Suspended() = 0;
			// printf("DTN (%d): Retrans bundle = %d, offset = %d, size = %d, for = %d, storedBundles = %d, RTT = %f\n", addr(), 
			//	pExpiredBundle->HeaderDTN().SequenceNumber(), pExpiredBundle->HeaderDTN().FragmentOffset(), 
			//	pExpiredBundle->HeaderDTN().BundleSize(), pExpiredBundle->Forwarded(), bundleManagerDTN_.BundlesStored(),
			//	pExpiredBundle->RetransmitTime());
			// pExpiredBundle->Forwarded() = 1;
			SendDataBundle(pExpiredBundle, pExpiredBundle->ReceivedBytes() + DTN_HEADER_SIZE);
		}
		else
			SetBundleRTTime(pExpiredBundle);
		
		pExpiredBundle = bundleManagerDTN_.GetFirstExpiredBundle(pExpiredBundle);
	}
	
	UpdateRTTimer();
}

/* Called when the rendezvous timer expires and sends all oustanding bundles. */
void DtnAgent::RVTimerHandler()
{
	double fSimulatorTime = Scheduler::instance().clock();
	double fReceptionInterval = fSimulatorTime - fLastFlushTime_;
	double fBufferPortion = (1.0) * m_nBytesReceived / nBDProduct_;
	double fNextTimeout;
	// double channelUtilization = 0;

	// if ( pMac_ )
	//	channelUtilization = pMac_->GetUtilization();
	
	double fSmoothedBufferPortion = fBufferPortion - ((fBufferPortion - 1) / 2);


	if ( fBufferPortion == 0 )
		fNextTimeout = fReceptionInterval * 2;
	else
		fNextTimeout = fReceptionInterval / fSmoothedBufferPortion;

	// Tcl& tcl = Tcl::instance();
	// tcl.evalf("[Simulator instance] puts-ns-traceall \"Buffer emptied %f (timeout) at %f. Next timeout in %f\"", 
		// nBytesReceived_/1000.0, fSimulatorTime, fNextTimeout);
	
	// printf("Base Station: Time = %f, Buffer emptied %d, Total bytes = %d, Channel utilization = %f, (old = %f)!\n", fSimulatorTime, nBytesReceived_,
		// nTotalBytesReceived_, channelUtilization, pMac_->GetUtilizationOld());
	SendOutstandingBundles();
	debug("Next timeout at %f!\n", fSimulatorTime + fNextTimeout);
	RVtimer_.resched(fNextTimeout);
	
	//printf("Base Station: Total bytes = %d!\n", nTotalBytesReceived_);
	fLastFlushTime_ = fSimulatorTime;
	m_nBytesReceived = 0;
}

/* Fragments all bundles to the bytes received so far and sends the fragments to the next hop (mobile node). */
void DtnAgent::SendOutstandingBundles()
{
	bundleManagerDTN_.FragmentAllBundles();
	Bundle *pUnsentBundle = bundleManagerDTN_.GetFirstUnSentBundle(NULL);
	while (pUnsentBundle)
	{
		debug("DTN (%d): Sending bundle = %d, fragment = %d, size = %d, for = %d, storedBundles = %d\n", addr(), 
			pUnsentBundle->HeaderDTN().SequenceNumber(), pUnsentBundle->HeaderDTN().FragmentOffset(), 
			pUnsentBundle->HeaderDTN().BundleSize(), pUnsentBundle->Forwarded(), bundleManagerDTN_.BundlesStored());
		
		pUnsentBundle->HeaderDTN().LeadingBytes() = 1;
		pUnsentBundle->Forwarded() = 1;
		
		SendDataBundle(pUnsentBundle, pUnsentBundle->ReceivedBytes() + DTN_HEADER_SIZE);
		pUnsentBundle = bundleManagerDTN_.GetFirstUnSentBundle(pUnsentBundle);
	}
}
